import unittest
from intelligen import numeric

def f(x): return x**3 + 2*x**2 + 10*x - 20
def df(x): return 3*x**2 + 4*x + 10
def ddf(x): return 6*x + 4
ROOT = 1.36880

class TestNumeric(unittest.TestCase):
    """Tests for `intelligen` package."""

    def test_newton(self):
        z = numeric.newton(f, df, 1, 0.01, True)[0]
        self.assertAlmostEqual(z, ROOT, 3)

    def test_bisection(self):
        z = numeric.bisection(f, 1, 2, 0.01, True)[0]
        self.assertAlmostEqual(z, ROOT, 3)

    def test_regula_falsi(self):
        z = numeric.regula_falsi(f, 1, 2, 0.01, True)[0]
        self.assertAlmostEqual(z, ROOT, 3)
    
    def test_secant(self):
        z = numeric.secant(f, 1, 2, 0.01, True)[0]
        self.assertAlmostEqual(z, ROOT, 3)
    
    def test_newton(self):
        z = numeric.newton2(f, df, ddf, 1, 0.01, True)[0]
        self.assertAlmostEqual(z, ROOT, 3)

if __name__ == '__main__':
    unittest.main()



import unittest
import numpy as np
import matplotlib.pyplot as plt

from intelligen import interpolate

X = np.array([-1, 0, 4,  1, 7, 8])
X = np.concatenate((X, X+10))
Y = np.array([ 4, 2, 3, -2, 6, 6])
Y = np.concatenate((Y, Y+1))

_x = np.linspace(min(X),max(X),1000)

Lagran = interpolate.lagrange(X.reshape(-1),Y.reshape(-1))

class TestInterpolate(unittest.TestCase):
    """Tests for `intelligen` package."""
    
    def test_lagrange(self):
        Lagran(_x)
    

if __name__ == '__main__':
    unittest.main()



import numpy as np
import matplotlib.pyplot as plt
from typing import List, Tuple, Callable, Union

Vector = List[float]
Function = Callable[[float], float]
Function2d = Callable[[float, float], float]

def newton_cotes(y: Vector=None, x: Vector=None, formula: str='trapz', h: float=1,
                 f: Function=None, a: float=None, b: float=None, N: int=100) -> float:
    """
    Newton-Cotes rules
    ==================
    In numerical analysis, the Newton-Cotes formulas, also called the Newton-Cotes quadrature
    rules or simply Newton-Cotes rules, are a group of formulas for numerical integration
    (also called quadrature) based on evaluating the integrand at equally spaced points.
    They are named after Isaac Newton and Roger Cotes.

    Newton-Cotes formulas can be useful if the value of the integrand at equally spaced points is given.
    If it is possible to change the points at which the integrand is evaluated,
    then other methods such as Gaussian quadrature and Clenshaw-Curtis quadrature are probably more suitable.
    `n` is the grade of the polynomial interpolation, also nodes - 1 

    Parameters
    ----------
    y : Vector
        Input array to integrate
    x : Vector, optional
        The sample points corresponding to the y values. If x is None,
        the sample points are assumed to be evenly spaced h apart, by default None
    formula : str, optional
        This defines which Newton-Cotes rule apply
            trapz: trapezoidal rule, `n=1`
            simpson: Simpson's rule, `n=2`
            simpson3_8: Simpson's rule 3/8, `n=3`
            boole : Boole's rule, `n=4`
            weddle: Weddle's rule, `n=6`

            -> For simplicity is recommended to use the functions `trapz`, `simpson` etc.
            
    h : float, optional
        The spacing between sample points when x is None, by default 1
    f : function, optional
        funcion to integrate, if you don't define y/x
    a : float, optional
        Lower limit of the integral
    b : float, optional
        Upper limit of the integral
    N : int, optional
        Number of intervals, must be divisible by n, by default 100

    Returns
    -------
    float
        Definite integral
    
    Examples
    --------
    >>> N, a, b = 12, 1, 4
    >>> x = np.linspace(a, b, N+1)
    >>> y = np.sin(x)
    >>> print(newton_cotes(y,x,'simpson'))
    1.193972031074762
    >>> def fun(x): return np.sin(x)
    >>> print(newton_cotes(f=fun, a=a, b=b, N=N, formula='simpson3_8'))
    1.1940051049177476

    """
    if   formula == 'trapz': 
        n, fs = 1, '(y[i] + y[i+1]) * d[i]/2'
    elif formula == 'simpson':
        n, fs = 2, '(y[n*i] + 4*y[n*i + 1] + y[n*i + 2]) * d[i]/3'
    elif formula == 'simpson3_8': 
        n, fs = 3, '(y[n*i] + 3*y[n*i + 1] + 3*y[n*i + 2] + y[n*i + 3]) * 3*d[i]/8'
    elif formula == 'boole': 
        n, fs = 4, '(7*y[n*i] + 32*y[n*i + 1] + 12*y[n*i + 2] + 32*y[n*i + 3] + 7*y[n*i + 4]) * 2*d[i]/45'
    elif formula == 'weddle': 
        n, fs = 6, '(y[n*i] + 5*y[n*i + 1] + y[n*i + 2] + 6*y[n*i + 3] + y[n*i + 4] + 5*y[n*i + 5] + y[n*i + 6]) * 3*d[i]/10'
    else: raise ValueError('Wrong formula')

    if y is None and x is None:
        if not all(p is not None for p in [f,a,b]): raise ValueError(f"If you don't define y/x you must define f,a and b")
        if N%n != 0: raise ValueError(f'{N=} must be divisible by {n=}')
        x = np.linspace(a,b,N+1)
        y = f(x)   
    
    if (len(y)-1)%n != 0: raise ValueError(f'The length of the array y={len(y)} - 1 must be multiple of {n=} in order to divide the interval')
    nh = int((len(y)-1)/n)

    if x is None: 
        d = np.array([h]*nh)
        # x is only for checking
        x = np.empty(len(y))
    else: d = np.diff(x)[::n]

    if len(y) != len(x): raise ValueError(f'The length of x={len(x)} and y={len(y)} must be the same')

    s = 0
    for i in range(nh):
        s += eval(fs)
    return s

def trapz(y: Vector=None, x: Vector=None, h: float=1,
          f: Function=None, a: float=None, b: float=None, N: int=100) -> float:
    """
    Trapezoidal Rule
    ================
    The trapezoidal rule (also known as the trapezoid rule or trapezium rule)
    is a technique for approximating the definite integral.
    The trapezoidal rule works by approximating the region under the graph of the function 
    `f(x)` as a trapezoid and calculating its area.
    `n=1`, where n is the grade of the polynomial interpolation, also nodes - 1.
    This is a special case of Newton-Cotes formulas  

    Parameters
    ----------
    y : Vector
        Input array to integrate
    x : Vector, optional
        The sample points corresponding to the y values. If x is None,
        the sample points are assumed to be evenly spaced h apart, by default None
    h : float, optional
        The spacing between sample points when x is None, by default 1
    f : function, optional
        funcion to integrate, if you don't define y/x
    a : float, optional
        Lower limit of the integral
    b : float, optional
        Upper limit of the integral
    N : int, optional
        Number of intervals, must be divisible by n, by default 100

    Returns
    -------
    float
        Definite integral
    
    Examples
    --------
    >>> N, a, b = 12, 1, 4
    >>> x = np.linspace(a, b, N+1)
    >>> y = np.sin(x)
    >>> print(trapz(y,x))
    1.187720971137812
    >>> def fun(x): return np.sin(x)
    >>> print(trapz(f=fun, a=a, b=b, N=N))
    1.187720971137812

    """
    return newton_cotes(y, x, 'trapz', h, f, a, b, N)

def simpson(y: Vector=None, x: Vector=None, h: float=1,
            f: Function=None, a: float=None, b: float=None, N: int=100) -> float:
    """
    Simpson's Rule
    ============
    In numerical integration, Simpson's rules are several approximations for definite integrals,
    named after Thomas Simpson (1710-1761).
    This is the  most basic of these rules, called Simpson's 1/3 rule, or just Simpson's rule.
    `n=2`, where n is the grade of the polynomial interpolation, also nodes - 1.
    This is a special case of Newton-Cotes formulas 

    Parameters
    ----------
    y : Vector
        Input array to integrate
    x : Vector, optional
        The sample points corresponding to the y values. If x is None,
        the sample points are assumed to be evenly spaced h apart, by default None
    h : float, optional
        The spacing between sample points when x is None, by default 1
    f : function, optional
        funcion to integrate, if you don't define y/x
    a : float, optional
        Lower limit of the integral
    b : float, optional
        Upper limit of the integral
    N : int, optional
        Number of intervals, must be divisible by n, by default 100

    Returns
    -------
    float
        Definite integral
    
    Examples
    --------
    >>> N, a, b = 12, 1, 4
    >>> x = np.linspace(a, b, N+1)
    >>> y = np.sin(x)
    >>> print(simpson(y,x))
    1.193972031074762
    >>> def fun(x): return np.sin(x)
    >>> print(simpson(f=fun, a=a, b=b, N=N))
    1.193972031074762

    """
    return newton_cotes(y, x, 'simpson', h, f, a, b, N)

def simpson3_8(y: Vector=None, x: Vector=None, h: float=1,
               f: Function=None, a: float=None, b: float=None, N: int=99) -> float:
    """
    Simpson's Rule 3/8
    ================
    In numerical integration, Simpson's rules are several approximations for definite integrals,
    named after Thomas Simpson (1710-1761).
    This is the second of these rules, called Simpson's 3/8 rule or Simpson's second rule.
    `n=3`, where n is the grade of the polynomial interpolation, also nodes - 1.
    This is a special case of Newton-Cotes formulas 

    Parameters
    ----------
    y : Vector
        Input array to integrate
    x : Vector, optional
        The sample points corresponding to the y values. If x is None,
        the sample points are assumed to be evenly spaced h apart, by default None
    h : float, optional
        The spacing between sample points when x is None, by default 1
    f : function, optional
        funcion to integrate, if you don't define y/x
    a : float, optional
        Lower limit of the integral
    b : float, optional
        Upper limit of the integral
    N : int, optional
        Number of intervals, must be divisible by n, by default 99

    Returns
    -------
    float
        Definite integral
    
    Examples
    --------
    >>> N, a, b = 12, 1, 4
    >>> x = np.linspace(a, b, N+1)
    >>> y = np.sin(x)
    >>> print(simpson3_8(y,x))
    1.1940051049177476
    >>> def fun(x): return np.sin(x)
    >>> print(simpson3_8(f=fun, a=a, b=b, N=N))
    1.1940051049177476

    """
    return newton_cotes(y, x, 'simpson3_8', h, f, a, b, N)

def boole(y: Vector=None, x: Vector=None, h: float=1,
          f: Function=None, a: float=None, b: float=None, N: int=100) -> float:
    """
    Boole's Rule 
    ============
    In mathematics, Boole's rule, named after George Boole, is a method of numerical integration.

    `n=4`, where n is the grade of the polynomial interpolation, also nodes - 1
    This is a special case of Newton-Cotes formulas 

    Parameters
    ----------
    y : Vector
        Input array to integrate
    x : Vector, optional
        The sample points corresponding to the y values. If x is None,
        the sample points are assumed to be evenly spaced h apart, by default None
    h : float, optional
        The spacing between sample points when x is None, by default 1
    f : function, optional
        funcion to integrate, if you don't define y/x
    a : float, optional
        Lower limit of the integral
    b : float, optional
        Upper limit of the integral
    N : int, optional
        Number of intervals, must be divisible by n, by default 100

    Returns
    -------
    float
        Definite integral

    """
    return newton_cotes(y, x, 'boole', h, f, a, b, N)

def weddle(y: Vector=None, x: Vector=None, h: float=1,
           f: Function=None, a: float=None, b: float=None, N: int=96) -> float:
    """
    Weddle's Rule 
    ==========

    n=6, where n is the grade of the polynomial interpolation, also nodes - 1 

    Parameters
    ----------
    y : Vector
        Input array to integrate
    x : Vector, optional
        The sample points corresponding to the y values. If x is None,
        the sample points are assumed to be evenly spaced h apart, by default None
    h : float, optional
        The spacing between sample points when x is None, by default 1
    f : function, optional
        funcion to integrate, if you don't define y/x
    a : float, optional
        Lower limit of the integral
    b : float, optional
        Upper limit of the integral
    N : int, optional
        Number of intervals, must be divisible by n, by default 96

    Returns
    -------
    float
        Definite integral

    """
    return newton_cotes(y, x, 'weddle', h, f, a, b, N)

def newton_cotes2(y: Vector=None, x: Vector=None, h: float=1,
                 f: Function=None, a: float=None, b: float=None, n: int=100) -> float:
    n = len(y)-1
    if   n == 1: return trapz(y,x,h)
    elif n == 2: return simpson(y,x,h)
    elif n == 3: return simpson3_8(y,x,h)
    elif n == 4: return boole(y,x,h)
    elif n == 6: return weddle(y,x,h)


def odeEuler(f: Function2d, t0: float, tfin: float, N: int, y0: Union[float, Vector]) -> Tuple[Vector, Vector]:
    """
    Euler Method
    ============

    Parameters
    ----------
    f : Function2d
        Right-hand side of the differential equation `y' = f(t,y), y(t0) = y0`
    t0 : float
        _description_
    tfin : float
        _description_
    N : int
        _description_
    y0 : float
        _description_

    Returns
    -------
    Tuple[Vector, Vector]
        _description_
    
    Examples
    --------
    >>> def q(t): return 1/2 + 1/2 * np.cos(t**2)
    >>> def f(t, y): return q(t) - y
    >>> T, U = odeEuler(f, 0, 6, 1000, 0)
    >>> import matplotlib.pyplot as plt
    >>> plt.plot(T,U)
    >>> plt.show()
    plot figure
    >>> T, U = odeEuler(f, 0, 6, 1000, [0,1])
    >>> for i in range(len(T)):
            plt.plot(T[i], U[i])
    >>> plt.show()
    plot figure

    """
    if isinstance(y0, list) or type(y0) == np.ndarray:
        leny0 = len(y0)
        
        T = np.tile(np.linspace(t0, tfin, N+1), (leny0,1))
        h = (tfin - t0) / N

        U = np.empty([leny0, N+1])
        U[:,0] = y0
        for i in range(1, N+1):
            U[:,i] = U[:,i-1] + h*f(T[:,i-1], U[:,i-1])

    else:
        T = np.linspace(t0, tfin, N+1)
        h = (tfin - t0) / N
        U = np.empty(N+1)
        U[0] = y0
        for i in range(1, N+1):
            U[i] = U[i-1] + h*f(T[i-1], U[i-1])

    return T, U

def slope_field(f: Function2d, range: list = None,
                xlim: list = None, ylim: list = None,
                normalize: bool = True, plot_type: str = 'quiver',
                density: int = 20, color: bool = True,
                show: bool = True, cmap: str = 'viridis'):
    """
    Slope Field
    ===========
    Graphical representation of the solutions
    to a first-order differential equation `y' = f(t,y)`

    Parameters
    ----------
    f : Function2d
        Function with 2 parameters `y' = f(t,y)`

    range : list, optional
        Sets both limits x/y of the plot, by default [-5, 5]

    xlim : list, optional
        Sets the x limits of the plot, if `range` is defined, xlim is already set, by default [-5, 5]

    ylim : list, optional
        Sets the y limits of the plot, if `range` is defined, ylim is already set, by default [-5, 5]

    normalize : bool, optional
        Normalize the slope field, by default True

    plot_type : str, optional
        Defines the plot type
            quiver: -> plt.quiver()
            streamplot: -> plt.streamplot()
        by default 'quiver'

    density : int, optional
        _description_, by default 20

    color : bool, optional
        _description_, by default True
        
    show : bool, optional
        _description_, by default True

    cmap : str, optional
        https://matplotlib.org/stable/tutorials/colors/colormaps.html

    Examples
    --------
    >>> def fun(x,y): return x + np.sin(y)
    >>> slope_field(fun, range=[-2,2], plot_type='streamplot', cmap='plasma')
    plot figure
    >>> slope_field(fun, xlim=[-3,2], ylim=[-1,1], color=False, normalize=False, density=30, show=False)
    >>> T, U = odeEuler(fun, -3, 2, 1000, 0.1)
    >>> import matplotlib.pyplot as plt
    >>> plt.plot(T,U)
    >>> plt.show()
    plot figure

    """
    
    if range is None and xlim is None and ylim is None:
        range = [-5,5]
        x1, x2 = range
        y1, y2 = range
    
    elif xlim is None and ylim is None:
        x1, x2 = range
        y1, y2 = range
    
    elif range is None:
        x1, x2 = xlim
        y1, y2 = ylim
    
    else:
        raise ValueError('Must speciefy either range or xlim/ylim')
    

    x = np.linspace(x1, x2, density)
    y = np.linspace(y1, y2, density)

    X, Y = np.meshgrid(x, y)

    dx, dy = np.ones(X.shape), f(X,Y)
    if normalize:
        norm = np.sqrt(dx**2 + dy**2)
        dx, dy = dx/norm , dy/norm
    
    if plot_type == 'quiver':
        #color = np.sqrt(((dx+4)/2)*2 + ((dy+4)/2)*2)
        if color: plt.quiver(X, Y, dx, dy, dy, cmap=cmap)
        else: plt.quiver(X, Y, dx, dy)

    elif plot_type == 'streamplot':
        if color: plt.streamplot(X, Y, dx, dy, color=dy, cmap=cmap)
        else: plt.streamplot(X, Y, dx, dy, color='k')

    else:
        raise ValueError("It only accepts either 'quiver' or 'streamplot'")
    
    plt.title(f'Slope Field ({plot_type})')
    plt.xlabel('x')
    plt.ylabel('y')
    if show: plt.show()


import numpy as np
import matplotlib.pyplot as plt
from typing import List
Vector = List[float]

def heaviside(t: Vector) -> Vector:
    return (np.array(t) >= 0).astype(int)

def rect(t: Vector) -> Vector:
    t = np.array(t)
    return heaviside(t + 1/2) - heaviside(t - 1/2)

def delta(n: Vector) -> Vector:
    return (np.array(n) == 0).astype(int)

def time_delay(x: Vector, y: Vector, t: float) -> Vector:
    x, y = np.array(x), np.array(y)
    x = x + t
    return x, y




"""Top-level package for intelligen."""

__author__ = """Diego Bouchet"""
__email__ = 'diegobouchet88@gmail.com'
__version__ = '0.12.2'

submodules = [
        'AI',
        'constants',
        'integrate',
        'intelligen',
        'interpolate',
        'linregress',
        'numeric',
        'signals',
        'stats'
    ]

__all__ = submodules

from . import AI
from . import constants
from . import integrate
from . import intelligen
from . import interpolate
from . import linregress
from . import numeric
from . import signals
from . import stats



import numpy as np
from intelligen.fortran.functions import erf as erf_fortran

from typing import List, Union
Vector = List[float]

__all__ = ['mean_squared_error', 'mean_absolute_error',
           'erf', 'erfc']

def mean_squared_error(y_real: Vector, y_pred: Vector) -> float:
    """Returns the mean squared error
    Args:
        y_real (Vector): Real data
        y_pred (Vector): Predicted data
    Returns:
        float: mean squared error
    """
    y_real = np.array(y_real)
    y_pred = np.array(y_pred)
    
    return np.mean((y_real - y_pred)**2)
def mean_absolute_error(y_real: Vector, y_pred: Vector) -> float:
    """Returns the mean absolute error
    Args:
        y_real (Vector): Real data
        y_pred (Vector): Predicted data
    Returns:
        float: mean absolute error
    """
    y_real = np.array(y_real)
    y_pred = np.array(y_pred)
    
    return np.mean(np.abs(y_real - y_pred))


def erf(x: Union[float, Vector]) -> Union[float, Vector]:
    """
    Error Function
    ==============

    Parameters
    ----------
    x : float, Vector
        Value

    Returns
    -------
    float, Vector
        Error function at x
    """
    
    try:
        return erf_fortran(x).item()
    except:
        return erf_fortran(x)

def erfc(x: Union[float, Vector]) -> Union[float, Vector]:
    """
    Complementary Error Function
    ==============

    Parameters
    ----------
    x : float, Vector
        Value

    Returns
    -------
    float, Vector
        Complementary error function at x
    """
    return 1 - erf(x)


import numpy as np
from typing import Callable, List, Union
Function = Callable[[float], float]
Vector_int = List[int]
from .constants import golden, igolden


def fibonacci(n: int, list: bool = False, start_points: Vector_int = None) -> int:
    """
    Fibonacci Numbers
    =================
    In mathematics, the Fibonacci numbers, commonly denoted Fn,
    form a sequence, the Fibonacci sequence, in which each number
    is the sum of the two preceding ones.

    Parameters
    ----------
    n : int
        nth element of the Fibonacci sequence
    list : bool, optional
        Shows the whole sequence until nth number, by default False
    start_points : Vector_int, optional
        Initial numbers of the sequence, by default [0, 1]

    Returns
    -------
    int
        nth element of the Fibonacci sequence
    """    """"""
    if start_points is None: start_points = [0, 1]
    f0, f1 = start_points

    if list:
        if n == 0: return [f0]
        if n == 1: return [f0, f1]

        F = start_points.copy()
        if n > 0:
            for i in range(1, n):
                F.append(F[i] + F[i-1])
        else:
            for _ in range(-n):
                F.insert(0, F[1] - F[0])

        return F
    else:
        if n == 0: return f0
        if n == 1: return f1

        if n > 0:
            for i in range(n-1):
                f1, f0 = f1 + f0, f1
            return f1
        else:
            for i in range(-n):
                f0, f1 = f1 - f0, f0
            return f0
            # Another way:
            #return (-1)**(n+1) * fibonacci(-n)
            
def binet(n: float):
    return (golden**n - igolden**n)/np.sqrt(5)

def factorial(n: Union[int, Vector_int]) -> Union[int, Vector_int]:
    """
    Factorial
    =========
    Returns the factorial of an integer `n! -> n*(n-1)*...*1`

    Parameters
    ----------
    n: int, Vector_int
        The number, or list of numbers to perform the factorial


    Returns
    -------
    int, Vector_int
        Factorial
    """
    if isinstance(n, int):
        result = 1
        for i in range(2, n+1):
            result *= i
        return result
    else:
        if type(n) is np.ndarray: n = list(n)    

        result = [1 for e in n]
        for i, e in enumerate(n):
            for j in range(2, e+1):
                result[i] *= j
  
        return np.asarray(result)

def combination(n: int, k: Union[int, Vector_int]) -> Union[int, Vector_int]:
    """
    Combination
    ===========
    Returns the number of combinations for a set of n items in k selected items

    Parameters
    ----------
    n: int
        total number of items
    k: int, Vector_int
        selected number of items

    Returns
    -------
    int, Vector_int
        number of combinations
    """
    if isinstance(k, list) or type(k) == np.ndarray:
        return np.array([combination(n,i) for i in k])
    else:
        if isinstance(k, int):
            return int(factorial(n) / (factorial(k)*factorial(n-k)))
        else:
            raise ValueError('k must be integer')


def derivative(f: Function ,a: float, order: int=1, method: str='central', h: float=0.01):
    '''
    Derivative
    ========
    Compute the difference formula for `f'(a)` with step size `h`.

    Parameters
    ----------
    f : Function
        Vectorized function of one variable, MUST accept arrays as input
    a : float
        Compute derivative at `x = a`
    method : string
        Difference formula: (order=1) 
            central: `f(a+h) - f(a-h))/2h`
            forward: `f(a+h) - f(a))/h`
            backward: `f(a) - f(a-h))/h`

            by default 'central'
        Information about higher order https://en.wikipedia.org/wiki/Finite_difference

    h : float
        Step size in difference formula,
        by default 0.01
    
    order: int
        Order of the derivative

    Returns
    -------
    float
        `f'(a)`      
    '''
    i = np.arange(order+1)
    if method == 'central':
        return 1/h**order * np.sum((-1)**i * combination(order, i) * f(a + (order/2 - i)*h))

    elif method == 'forward':
        return  1/h**order * np.sum((-1)**(order - i) * combination(order, i) * f(a + i*h))

    elif method == 'backward':
        return 1/h**order * np.sum((-1)**i * combination(order, i) * f(a - i*h))

    else:
        raise ValueError("Method must be 'central', 'forward' or 'backward'.")


#---------------Find Root---------------#


def newton(f: Function, x: float, tol: float, iter: bool = False) -> float:
    """
    Newton
    ======
    Newton method to find a root of a function

    Parameters
    ----------
    f : Function
        Function
    x : float
        Start point
    tol : float
        Error tolerance
    iter : bool, optional
        Shows the iterations needed to find the solution,
        by default False

    Returns
    -------
    float
        The root
    int, optional
        Number of iterations
    
    Examples
    --------
    >>> def f(x): return x**3 + 2*x**2 + 10*x - 20
    >>> newton(f, 1, 0.01, True)
    (1.3688081886175318, 3)


    """
    n = 0
    while abs(f(x)) > tol:
        x = x - f(x) / derivative(f, x, h=1e-5)
        n += 1

    if iter:
        return x, n
    return x

def bisection(f: Function, xi: float, xf: float, tol: float, iter: bool = False) -> float:
    """
    Bisection
    =========
    Bisection method to find a root of a function

    Parameters
    ----------
    f : Function
        Function
    xi : float
        First point
    xf : float
        Second point
    tol : float
        Error tolerance
    iter : bool, optional
        Shows the iterations needed to find the solution,
        by default False

    Returns
    -------
    float
        The root
    int, optional
        Number of iterations
    
    Examples
    --------
    >>> def f(x): return x**3 + 2*x**2 + 10*x - 20
    >>> bisection(f, 1, 2, 0.01, True)
    (1.369140625, 9)

    """
    if f(xi) * f(xf) < 0:
        xm, n = (xi + xf) / 2, 1

        while abs(f(xm)) > tol:
            if f(xi) * f(xm) < 0:
                xf = xm 
                n += 1
            
            elif f(xm) * f(xf) < 0:
                xi = xm 
                n += 1
            
            xm = (xi + xf) / 2
            
        if iter:
            return xm, n
        return xm

    else:
        print("Invalid input")

def regula_falsi(f: Function, xi: float, xf: float, tol: float, iter: bool = False) -> float:
    """
    Regula Falsi
    ============
    Regula falsi method to find a root of a function
    

    Parameters
    ----------
    f : Function
        Function
    xi : float
        First point
    xf : float
        Second point
    tol : float
        Error tolerance
    iter : bool, optional
        Shows the iterations needed to find the solution,
        by default False

    Returns
    -------
    float
        The root
    int, optional
        Number of iterations
    
    Examples
    --------
    >>> def f(x): return x**3 + 2*x**2 + 10*x - 20
    >>> regula_falsi(f, 1, 2, 0.01, True)
    (1.3685009755999702, 4)

    """
    if f(xi) * f(xf) < 0:
        xm, n = (xi * f(xf) - xf * f(xi)) / (f(xf) - f(xi)), 1

        while abs(f(xm)) > tol:
            if f(xi) * f(xm) < 0:
                xf = xm
                n += 1

            if f(xm) * f(xf) < 0:
                xi = xm
                n += 1
            
            xm = (xi * f(xf) - xf * f(xi)) / (f(xf) - f(xi))

        if iter:
            return xm, n
        return xm
    else:
        print("Invalid input")


def secant(f: Function, x0: float, x1: float, tol: float, iter: bool = False) -> float:
    """
    Secant
    ======

    Secant method to find a root of a function

    Parameters
    ----------
    f : Function
        Function
    x0 : float
        First point
    x1 : float
        Second point
    tol : float
        Error tolerance
    iter : bool, optional
        Shows the iterations needed to find the solution,
        by default False

    Returns
    -------
    float
        The root
    int, optional
        Number of iterations
    
    Examples
    --------
    >>> def f(x): return x**3 + 2*x**2 + 10*x - 20
    >>> secant(f, 1, 2, 0.01, True)
    (1.369013325992566, 3)

    """
    x2, n = x1 - (f(x1) * (x1 - x0)) / (f(x1) - f(x0)), 1

    while abs(f(x2)) > tol:
        x0, x1 = x1, x2
        x2 = x1 - (f(x1) * (x1 - x0)) / (f(x1) - f(x0))
        n += 1
        
    if iter:
        return x2, n
    return x2



""" def fixed_point(g, dg, x, tol, iter = False):
  ''' f(x) = 0 ==> x = g(x) '''
  n = 1
  if abs(dg(x))<1:
    xa = x
    x = g(x)
    
    while abs(x-xa)>tol:
      xa = x
      x = g(x)
      n = n+1
    if iter: return x, n
    return x
  else:
    print("Doesn't converge")  """


def newton2(f: Function, x: float, tol: float, iter: bool = False) -> float:
    """
    Newton 2nd order
    ================

    Newton second order method to find a root of a function

    Parameters
    ----------
    f : Function
        Function
    x : float
        Start point
    tol : float
        Error tolerance
    iter : bool, optional
        Shows the iterations needed to find the solution,
        by default False

    Returns
    -------
    float
        The root
    int, optional
        Number of iterations
    
    Examples
    --------
    >>> def f(x): return x**3 + 2*x**2 + 10*x - 20
    >>> newton2(f, 1, 0.01, True)
    (1.3688081071467233, 2)

    """
    n = 0

    while abs(f(x)) > tol:
        fx, dfx, ddfx = f(x), derivative(f, x), derivative(f, x, 2)

        x1 = x - dfx / ddfx + np.sqrt(dfx**2 - 2*ddfx * fx) / ddfx
        x2 = x - dfx / ddfx - np.sqrt(dfx**2 - 2*ddfx * fx) / ddfx

        if abs(f(x1)) < abs(f(x2)):
            x = x1
        else:
            x = x2
        
        n += 1
    
    if iter:
        return x, n
    return x



"""
Collection of physical constants and conversion factors
"""

import numpy as np

__all__ = ['golden', 'golden_ratio', 'igolden', 'inverse_golden_ratio',
           'G', 'g', 'c', 'speed_of_light', 'yotta', 'zetta', 'exa',
           'peta', 'tera', 'giga', 'mega', 'mega', 'kilo', 'hecto',
           'deka', 'deci', 'centi', 'milli', 'micro', 'nano', 'pico',
           'femto', 'atto', 'zepto', 'yocto']

golden = golden_ratio = (1 + np.sqrt(5))/2
igolden = inverse_golden_ratio = (1 - np.sqrt(5))/2

G = 6.67430e-11
g = 9.807
c = speed_of_light = 299_792_458

# SI prefixes
yotta = 1e24
zetta = 1e21
exa = 1e18
peta = 1e15
tera = 1e12
giga = 1e9
mega = 1e6
kilo = 1e3
hecto = 1e2
deka = 1e1
deci = 1e-1
centi = 1e-2
milli = 1e-3
micro = 1e-6
nano = 1e-9
pico = 1e-12
femto = 1e-15
atto = 1e-18
zepto = 1e-21
yocto = 1e-24


import numpy as np
import numpy.polynomial as pol
from typing import Callable, List, Union
Vector = List[float]
Function = Callable[[float], float]
Polynomial = pol.Polynomial


def lagrange2(X: Vector, Y: Vector) -> Function:
    """Returns the lagrange interpolation
    Args:
        X (Vector): X-data
        Y (Vector): Y-data
    Returns:
        Callable[[Vector], Vector]: returns the function that evaluates the lagrange interpolation
    """
    X = np.asarray(X)  

    def lagran(x):
        # Checks the input's type
        if type(x) is np.ndarray:
            x = x.reshape(-1, 1)
        
        if isinstance(x, list):
            x = np.array(x).reshape(-1, 1)
        
        else:
            x = np.array([x]).reshape(-1, 1)

        out = 0
        for i in range(len(X)):
            #pi_x = (x - np.array([(X[X != xi]).reshape(-1)] * len(x))) because X[X != xi] autoreshape (-1)
            pi_x = x - np.array([(X[X != X[i]])] * len(x))

            out += Y[i] * (pi_x / (X[i] - X[X != X[i]])).prod(axis = 1)

        return out
    
    return lagran
    
def lagrange(X: Vector, Y: Vector) -> Polynomial:
    p = 0
    for i in range(len(X)):
        qn = pol.Polynomial.fromroots(X[X != X[i]])
        p += Y[i] * qn / qn(X[i])
    return p
    


import numpy as np
import matplotlib.pyplot as plt
from ..fortran.functions import erf
#from math import erf
from typing import List, Union
Vector = List[float]

from numpy import pi, sqrt
from ..numeric import factorial, combination

__all__ = ['Distribution', 'DiscreteDistribution',
           'Bernoulli', 'Binomial', 'Geometric',
           'NegativeBinomial', 'Hypergeometric',
           'Poisson', 'Normal']

#---------------------- Distributions -----------------------#

class Distribution:
    """Main class for plotting distributions
    """


    def plot_cdf(self, show = True) -> None:
        """Plots the Cumulative distribution function of a distribution

        Args:
            show (bool, optional): Shows the plot or keep editing it. Defaults to True.
        """
        dist_name = self.__class__.__name__
        if not hasattr(self, 'CDF'):
            print(f'{dist_name} distribution doesnt support CDF plotting')
            return None
        plt.title(f'Cumulative distribution function\n{dist_name}')
        data = []
        for i in range(self.plot + 1):
            data.append(self.CDF(i))
        plt.step(range(len(data)), data, where='post')
        if show: plt.show()
    
    def plot_PDF(self, show = True) -> None:
        """Plots the Probability density function of a distribution

        Args:
            show (bool, optional): Shows the plot or keep editing it. Defaults to True.
        """
        dist_name = self.__class__.__name__
        if not hasattr(self, 'PDF'):
            print(f'{dist_name} distribution doesnt support PDF plotting')
            return None
        plt.title(f'Probability density function\n{dist_name}')
        data = []
        iter = np.linspace(-self.plot + self.origin, self.plot + self.origin, 1000)
        for i in iter:
            data.append(self.PDF(i))
        plt.plot(iter, data)
        if show: plt.show()


#------------------- Discrete Distributions --------------------#

class DiscreteDistribution(Distribution):
    
    def pmf(self, k: float) -> float:
        """
        Probability Mass Function
        =========================
        Returns the probability mass function P(x=k)

        Parameters
        ----------
        k : float
            Value
        
        Returns
        -------
        float:
            Probability
        """
    def CDF(k: float) -> float:
        """
        Cumulative Distribution Function
        ================================
        Returns the Cumulative distribution function P(x<=k)

        Parameters
        ----------
        k : float
            Value
        
        Returns
        -------
        float:
            Cumulative probability
        """

    def plot_pmf(self, show = True) -> None:
        """Plots the Probability mass function of a distribution

        Args:
            show (bool, optional): Shows the plot or keep editing it. Defaults to True.
        """
        dist_name = self.__class__.__name__
        if not hasattr(self, 'PMF'):
            print(f'{dist_name} distribution doesnt support PMF plotting')
            return None
        plt.title(f'Probability mass function\n{dist_name}')
        data = []
        for i in range(self.plot + 1):
            data.append(self.pmf(i))
        plt.plot(data)
        if show: plt.show()

class Bernoulli(DiscreteDistribution):
    """
    Bernoulli Distribution
    ======================

    Discrete probability distribution of a random variable
    which takes the value 1 with probability`p`and the
    value 0 with probability `q = 1-p`

    Parameters
    ----------
    p : float
        Probability of success
    
    Attributes
    ----------
    q : float
        `1-p`

    mean : float
        Expected value / mean 

    variance : float
        Measure of dispersion

    skewness : float
        Measure of the asymmetry

    kurtosis: float
        Measure of the "tailedness"

    Examples
    --------
    >>> from intelligen.stats import Bernoulli
    >>> B = Bernoulli(0.65)
    >>> B.pmf(0)
    0.35
    >>> B.skewness
    -0.6289709020331511

    """
    
    def __init__(self, p: float) -> None:
        self.p, self.q = p, 1-p
        self.mean = p
        self.variance = p * self.q
        self.skewness = (self.q - p) / np.sqrt(self.variance)
        self.kurtosis = (1 - 6*self.variance) / self.variance
    
    def __add__(self, distribution: Distribution) -> Distribution:
        if isinstance(distribution, self.__class__):
            if self.p == distribution.p:
                return Binomial(2, self.p)
            else: raise ValueError('Probability must be the same')
        
        elif isinstance(distribution, Binomial):
            if self.p == distribution.p:
                return Binomial(distribution.n + 1, self.p)
            else: raise ValueError('Probability must be the same')
        
        else: raise ValueError('Distributions must be compatible')
    
    def __radd__(self, distribution: Distribution) -> Distribution:
        return self + distribution
    
    def __mul__(self, coef: int) -> Distribution:
        if isinstance(coef, int) : return Binomial(coef, self.p)
        else: raise ValueError('The coeficient must be an integer')
    
    def __rmul__(self, coef: int) -> Distribution:
        return self * coef
    
    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(p={self.p})'
        


    def pmf(self, k: int) -> float:
        if isinstance(k, list) or type(k) == np.ndarray:
            return np.array([self.pmf(i) for i in k])
        else:
            if   k == 0: return self.q
            elif k == 1: return self.p
            else: raise ValueError('R{0,1}')

    def cdf(self, k: float) -> float:
        if isinstance(k, list) or type(k) == np.ndarray:
            return np.array([self.cdf(i) for i in k])
        else:
            if   k < 0: return 0
            elif k < 1: return self.q
            else: return 1

    def plot_pmf(self, show=True):
        plt.title(f'Probability mass function\nBernoulli')
        bar_plot = plt.bar([0,1],self.pmf([1,0]), width=0.4)
        plt.bar_label(bar_plot,['p','1-p'])
        if show: plt.show()
    
    def plot_cdf(self, show=True):
        plt.title(f'Cumulative distribution function\nBernoulli')
        plt.step([-1,0,1,2],self.cdf([-1,0,1,2]), where='post')
        if show: plt.show()

class Binomial(DiscreteDistribution):
    """
    Binomial Distribution
    =====================

    Discrete probability distribution that models the number
    of successes in a sequence of `n` independent experiments
    (Bernoulli trials) with constant probability `p`

    Parameters
    ----------
    n: int
        Number of Bernoulli trials
    
    p : float
        Probability of success

    Attributes
    ----------
    q : float
        `1-p`

    mean : float
        Expected value / mean 

    variance : float
        Measure of dispersion

    skewness : float
        Measure of the asymmetry

    kurtosis: float
        Measure of the "tailedness"

    Examples
    --------
    >>> from intelligen.stats import Binomial
    >>> B = Binomial(17, 0.65)
    >>> B.pmf(10)
    0.1684553555671748
    >>> B.kurtosis
    -0.09437621202327084
    """
    
    def __init__(self, n: int, p: float) -> None:
        self.n, self.p, self.q = n, p, 1-p
        self.plot = n
        self.mean = n * p
        self.variance = n * p * self.q
        self.skewness = (self.q - p) / np.sqrt(self.variance)
        self.kurtosis = (1 - 6*p*self.q) / self.variance
    
    def __add__(self, distribution: Distribution) -> Distribution:
        if isinstance(distribution, self.__class__):
            if self.p == distribution.p:
                return Binomial(self.n + distribution.n, self.p)
            else: raise ValueError('Probability must be the same')
        
        else: raise ValueError('Distributions must be compatible')

    def __mul__(self, coef: int) -> Distribution:
        if isinstance(coef, int) : return Binomial(self.n * coef, self.p)
        else: raise ValueError('The coeficient must be an integer')
    
    def __rmul__(self, coef: int) -> Distribution:
        return self * coef
    
    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(n={self.n}, p={self.p})'

    def pmf(self, k: int) -> float:
        if isinstance(k, list) or type(k) == np.ndarray:
            k = np.asarray(k).astype(int)
            print(k)
            #return np.array([self.pmf(i) for i in k])
        
        return combination(self.n, k) * self.p**k * (1 - self.p)**(self.n - k) 

    def cdf(self, k: float) -> float:
        if isinstance(k, list) or type(k) == np.ndarray:
            return np.array([self.cdf(i) for i in k])
        else:
            result = 0
            for i in range(int(np.floor(k))+1):
                result += combination(self.n, i) * self.p**i * (1 - self.p)**(self.n - i)
            return result

""" B = Binomial(10,0.5)
print(B.pmf([5,4])) """

class Geometric(DiscreteDistribution):
    """
    Geometric Distribution
    ======================

    Discrete probability distribution that models the number of
    Bernoulli trials `k`, needed to get one success, 
    each with success probability ``p`


    Parameters
    ----------
    p : float
        Probability of success

    Attributes
    ----------
    mean : float
        Expected value / mean 

    variance : float
        Measure of dispersion

    skewness : float
        Measure of the asymmetry

    kurtosis: float
        Measure of the "tailedness"

    Examples
    --------
    >>> from intelligen.stats import Geometric
    >>> G = Geometric(0.65)
    >>> G.CDF(3)
    0.957125
    >>> G.variance
    0.8284023668639052
    """
    def __init__(self, p: float, plot: int = None) -> None:
        self.p = p
        if plot is None: self.plot = int(10/p)
        self.mean = 1 / p
        self.variance = (1 - p) / p**2
        self.skewness = (2 - p) / np.sqrt(1 - p)
        self.kurtosis = 6 + p**2 / (1 - p)

    def __add__(self, distribution: Distribution) -> Distribution:
        if isinstance(distribution, self.__class__):
            if self.p == distribution.p:
                return NegativeBinomial(2, self.p)
            else: raise ValueError('Probability must be the same')
        
        elif isinstance(distribution, NegativeBinomial):
            if self.p == distribution.p:
                return NegativeBinomial(distribution.r + 1, self.p)
            else: raise ValueError('Probability must be the same')
        
        else: raise ValueError('Distributions must be compatible')

    def __radd__(self, distribution: Distribution) -> Distribution:
        return self + distribution

    def __mul__(self, coef: int) -> Distribution:
        if isinstance(coef, int) : return NegativeBinomial(coef, self.p)
        else: raise ValueError('The coeficient must be an integer')
    
    def __rmul__(self, coef: int) -> Distribution:
        return self * coef
    
    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(p={self.p})'

    def pmf(self, k: int or Vector) -> float:
        if isinstance(k, int):
            return (1 - self.p)**(k - 1) * self.p
        return (1 - self.p)**(np.array(k) - 1) * self.p
    
    def CDF(self, k: float) -> float:
        return 1 - (1 - self.p)**np.floor(k)

class NegativeBinomial(DiscreteDistribution):
    """
    Negative Binomial Distribution
    =====================

    Discrete probability distribution that models the number
    of successes in a sequence of independent and identically
    distributed Bernoulli trials before a specified
    (non-random) number of failures (denoted r) occurs

    Parameters
    ----------
    r: int
        Number of successes
    
    p : float
        Probability of success

    Attributes
    ----------
    q : float
        `1-p`

    mean : float
        Expected value / mean 

    variance : float
        Measure of dispersion

    skewness : float
        Measure of the asymmetry

    kurtosis: float
        Measure of the "tailedness"

    Examples
    --------
    >>> from intelligen.stats import Binomial
    >>> B = Binomial(17, 0.65)
    >>> B.PMF(10)
    0.1684553555671748
    >>> B.kurtosis
    -0.09437621202327084
    """
    def __init__(self, r: int, p: float, plot: int = None) -> None:
        self.r, self.p = r, p
        if plot is None: self.plot = int((2*(r+4)/(1-self.p)))
        self.mean = p*r / (1 - p)
        self.variance = p*r / (1 - p)**2
        self.skewness = (1 + p) / np.sqrt(p*r)
        self.kurtosis = 6/r + (1 - p)**2 / (p*r)

    def __add__(self, distribution: Distribution) -> Distribution:
        if isinstance(distribution, self.__class__):
            if self.p == distribution.p:
                return NegativeBinomial(self.r + distribution.r, self.p)
            else: raise ValueError('Probability must be the same')
        
        else: raise ValueError('Distributions must be compatible')

    def __mul__(self, coef: int) -> Distribution:
        if isinstance(coef, int) : return NegativeBinomial(self.r * coef, self.p)
        else: raise ValueError('The coeficient must be an integer')
    
    def __rmul__(self, coef: int) -> Distribution:
        return self * coef
    
    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(r={self.r}, p={self.p})'

    def pmf(self, k: int) -> float:
        return combination(k + self.r - 1, k) * (1 - self.p)**self.r * self.p**k
    
    def CDF(self, k: float) -> float:
        Binomial(k + self.r, self.p).CDF(k)
class Hypergeometric(DiscreteDistribution):
    
    def __init__(self, N: int, n: int, r: int) -> None:
        self.N, self.n, self.r, self.plot = N, n, r, r

    def pmf(self, k: int) -> float:
        return combination(self.r, k) * combination(self.N - self.r, self.n - k) / combination(self.N, self.n)


class Poisson(DiscreteDistribution):

    def __init__(self, l: float, plot: float = None) -> None:
        self.l = l
        if plot is None: self.plot = int(l*2) + 5
    
    def pmf(self, k: int) -> float:
        return np.exp(-self.l) * self.l**k / factorial(k)

class Normal(Distribution):

    def __init__(self, mu: float = 0, s: float = 1) -> None:
        self.mu, self.s, self.plot, self.origin = mu, s, 3*(s+1), mu
    
    def pmf(self, k: float) -> float:
        return np.exp((-1/2) * ((k - self.mu)/self.s)**2) / (self.s * np.sqrt(2*np.pi))
    
    def CDF(self, x: float) -> float:
        return 1/2 * (1 + erf((x - self.mu)/(self.s*np.sqrt(2))))
